@@ -9,6 +9,7 @@ use std::{
     sync::Arc,
 };
 
+use filesize::PathExt;
 use lmdb::DatabaseFlags;
 use log::LevelFilter;
 
@@ -22,8 +23,9 @@ use casper_execution_engine::{
             execution_result::ExecutionResult,
             run_genesis_request::RunGenesisRequest,
             step::{StepRequest, StepSuccess},
-            BalanceResult, EngineConfig, EngineState, GenesisSuccess, GetBidsRequest, QueryRequest,
-            QueryResult, StepError, SystemContractRegistry, UpgradeConfig, UpgradeSuccess,
+            BalanceResult, EngineConfig, EngineState, Error, GenesisSuccess, GetBidsRequest,
+            QueryRequest, QueryResult, StepError, SystemContractRegistry, UpgradeConfig,
+            UpgradeSuccess,
         },
         execution,
     },
@@ -36,10 +38,11 @@ use casper_execution_engine::{
     },
     storage::{
         global_state::{
-            in_memory::InMemoryGlobalState, lmdb::LmdbGlobalState, StateProvider, StateReader,
+            in_memory::InMemoryGlobalState, lmdb::LmdbGlobalState, scratch::ScratchGlobalState,
+            CommitProvider, StateProvider, StateReader,
         },
         transaction_source::lmdb::LmdbEnvironment,
-        trie::merkle_proof::TrieMerkleProof,
+        trie::{merkle_proof::TrieMerkleProof, Trie},
         trie_store::lmdb::LmdbTrieStore,
     },
 };
@@ -96,14 +99,12 @@ pub struct WasmTestBuilder<S> {
     genesis_account: Option<Account>,
     /// Genesis transforms
     genesis_transforms: Option<AdditiveMap<Key, Transform>>,
-    /// Mint contract key
-    mint_contract_hash: Option<ContractHash>,
-    /// Handle payment contract key
-    handle_payment_contract_hash: Option<ContractHash>,
-    /// Standard payment contract key
-    standard_payment_hash: Option<ContractHash>,
-    /// Auction contract key
-    auction_contract_hash: Option<ContractHash>,
+    /// Scratch global state used for in-memory execution and commit optimization.
+    scratch_engine_state: Option<EngineState<ScratchGlobalState>>,
+    /// System contract registry
+    system_contract_registry: Option<SystemContractRegistry>,
+    /// Global state dir, for implementations that define one.
+    global_state_dir: Option<PathBuf>,
 }
 
 impl<S> WasmTestBuilder<S> {
@@ -130,10 +131,9 @@ impl Default for InMemoryWasmTestBuilder {
             transforms: Vec::new(),
             genesis_account: None,
             genesis_transforms: None,
-            mint_contract_hash: None,
-            handle_payment_contract_hash: None,
-            standard_payment_hash: None,
-            auction_contract_hash: None,
+            system_contract_registry: None,
+            global_state_dir: None,
+            scratch_engine_state: None,
         }
     }
 }
@@ -151,10 +151,9 @@ impl<S> Clone for WasmTestBuilder<S> {
             transforms: self.transforms.clone(),
             genesis_account: self.genesis_account.clone(),
             genesis_transforms: self.genesis_transforms.clone(),
-            mint_contract_hash: self.mint_contract_hash,
-            handle_payment_contract_hash: self.handle_payment_contract_hash,
-            standard_payment_hash: self.standard_payment_hash,
-            auction_contract_hash: self.auction_contract_hash,
+            scratch_engine_state: None,
+            system_contract_registry: self.system_contract_registry.clone(),
+            global_state_dir: self.global_state_dir.clone(),
         }
     }
 }
@@ -164,14 +163,14 @@ impl InMemoryWasmTestBuilder {
     pub fn new(
         global_state: InMemoryGlobalState,
         engine_config: EngineConfig,
-        post_state_hash: Digest,
+        maybe_post_state_hash: Option<Digest>,
     ) -> Self {
         Self::initialize_logging();
         let engine_state = EngineState::new(global_state, engine_config);
         WasmTestBuilder {
             engine_state: Rc::new(engine_state),
-            genesis_hash: Some(post_state_hash),
-            post_state_hash: Some(post_state_hash),
+            genesis_hash: maybe_post_state_hash,
+            post_state_hash: maybe_post_state_hash,
             ..Default::default()
         }
     }
@@ -203,6 +202,7 @@ impl LmdbWasmTestBuilder {
 
         let global_state =
             LmdbGlobalState::empty(environment, trie_store).expect("should create LmdbGlobalState");
+
         let engine_state = EngineState::new(global_state, engine_config);
         WasmTestBuilder {
             engine_state: Rc::new(engine_state),
@@ -213,10 +213,9 @@ impl LmdbWasmTestBuilder {
             transforms: Vec::new(),
             genesis_account: None,
             genesis_transforms: None,
-            mint_contract_hash: None,
-            handle_payment_contract_hash: None,
-            standard_payment_hash: None,
-            auction_contract_hash: None,
+            system_contract_registry: None,
+            global_state_dir: Some(global_state_dir),
+            scratch_engine_state: None,
         }
     }
 
@@ -267,6 +266,7 @@ impl LmdbWasmTestBuilder {
 
         let global_state =
             LmdbGlobalState::empty(environment, trie_store).expect("should create LmdbGlobalState");
+
         let engine_state = EngineState::new(global_state, engine_config);
         WasmTestBuilder {
             engine_state: Rc::new(engine_state),
@@ -277,10 +277,9 @@ impl LmdbWasmTestBuilder {
             transforms: Vec::new(),
             genesis_account: None,
             genesis_transforms: None,
-            mint_contract_hash: None,
-            handle_payment_contract_hash: None,
-            standard_payment_hash: None,
-            auction_contract_hash: None,
+            scratch_engine_state: None,
+            system_contract_registry: None,
+            global_state_dir: Some(global_state_dir.as_ref().to_path_buf()),
         }
     }
 
@@ -298,11 +297,92 @@ impl LmdbWasmTestBuilder {
         path.push(GLOBAL_STATE_DIR);
         path
     }
+
+    /// Returns the file size on disk of the backing lmdb file behind LmdbGlobalState.
+    pub fn lmdb_on_disk_size(&self) -> Option<u64> {
+        if let Some(path) = self.global_state_dir.as_ref() {
+            let mut path = path.clone();
+            path.push("data.lmdb");
+            return path.as_path().size_on_disk().ok();
+        }
+        None
+    }
+
+    /// Execute and commit transforms from an ExecuteRequest into a scratch global state.
+    /// You MUST call write_scratch_to_lmdb to flush these changes to LmdbGlobalState.
+    pub fn scratch_exec_and_commit(&mut self, mut exec_request: ExecuteRequest) -> &mut Self {
+        if self.scratch_engine_state.is_none() {
+            self.scratch_engine_state = Some(self.engine_state.get_scratch_engine_state());
+        }
+
+        let cached_state = self
+            .scratch_engine_state
+            .as_ref()
+            .expect("scratch state should exist");
+
+        // Scratch still requires that one deploy be executed and committed at a time.
+        let exec_request = {
+            let hash = self.post_state_hash.expect("expected post_state_hash");
+            exec_request.parent_state_hash = hash;
+            exec_request
+        };
+
+        let mut exec_results = Vec::new();
+        // First execute the request against our scratch global state.
+        let maybe_exec_results = cached_state.run_execute(CorrelationId::new(), exec_request);
+        for execution_result in maybe_exec_results.unwrap() {
+            let journal = execution_result.execution_journal().clone();
+            let transforms: AdditiveMap<Key, Transform> = journal.clone().into();
+            let _post_state_hash = cached_state
+                .apply_effect(
+                    CorrelationId::new(),
+                    self.post_state_hash.expect("requires a post_state_hash"),
+                    transforms,
+                )
+                .expect("should commit");
+
+            // Save transforms and execution results for WasmTestBuilder.
+            self.transforms.push(journal.into());
+            exec_results.push(Rc::new(execution_result))
+        }
+        self.exec_results.push(exec_results);
+        self
+    }
+
+    /// Commit scratch to global state, and reset the scratch cache.
+    pub fn write_scratch_to_db(&mut self) -> &mut Self {
+        let prestate_hash = self.post_state_hash.expect("Should have genesis hash");
+        if let Some(scratch) = self.scratch_engine_state.take() {
+            let new_state_root = self
+                .engine_state
+                .write_scratch_to_db(prestate_hash, scratch.into_inner())
+                .unwrap();
+            self.post_state_hash = Some(new_state_root);
+        }
+        self
+    }
+
+    /// run step against scratch global state.
+    pub fn step_with_scratch(&mut self, step_request: StepRequest) -> &mut Self {
+        if self.scratch_engine_state.is_none() {
+            self.scratch_engine_state = Some(self.engine_state.get_scratch_engine_state());
+        }
+
+        let cached_state = self
+            .scratch_engine_state
+            .as_ref()
+            .expect("scratch state should exist");
+
+        cached_state
+            .commit_step(CorrelationId::new(), step_request)
+            .expect("unable to run step request against scratch global state");
+        self
+    }
 }
 
 impl<S> WasmTestBuilder<S>
 where
-    S: StateProvider,
+    S: StateProvider + CommitProvider,
     engine_state::Error: From<S::Error>,
     S::Error: Into<execution::Error>,
 {
@@ -329,13 +409,15 @@ where
         let genesis_account =
             utils::get_account(&transforms, &system_account).expect("Unable to get system account");
 
-        let registry = match self.query(
+        self.system_contract_registry = match self.query(
             Some(post_state_hash),
             Key::SystemContractRegistry,
             &empty_path,
         ) {
             Ok(StoredValue::CLValue(cl_registry)) => {
-                CLValue::into_t::<SystemContractRegistry>(cl_registry).unwrap()
+                let system_contract_registry =
+                    CLValue::into_t::<SystemContractRegistry>(cl_registry).unwrap();
+                Some(system_contract_registry)
             }
             Ok(_) => panic!("Failed to get system registry"),
             Err(err) => panic!("{}", err),
@@ -343,19 +425,6 @@ where
 
         self.genesis_hash = Some(post_state_hash);
         self.post_state_hash = Some(post_state_hash);
-        self.mint_contract_hash = Some(*registry.get(MINT).expect("should have mint hash"));
-        self.handle_payment_contract_hash = Some(
-            *registry
-                .get(HANDLE_PAYMENT)
-                .expect("should have handle payment hash"),
-        );
-        self.standard_payment_hash = Some(
-            *registry
-                .get(STANDARD_PAYMENT)
-                .expect("should have standard payment hash"),
-        );
-        self.auction_contract_hash =
-            Some(*registry.get(AUCTION).expect("should have auction hash"));
         self.genesis_account = Some(genesis_account);
         self.genesis_transforms = Some(transforms);
         self
@@ -431,7 +500,8 @@ where
     /// Panics if the total supply can't be found.
     pub fn total_supply(&self, maybe_post_state: Option<Digest>) -> U512 {
         let mint_key: Key = self
-            .mint_contract_hash
+            .get_system_contract_hash(MINT)
+            .cloned()
             .expect("should have mint_contract_hash")
             .into();
 
@@ -453,6 +523,7 @@ where
             exec_request.parent_state_hash = hash;
             exec_request
         };
+
         let maybe_exec_results = self
             .engine_state
             .run_execute(CorrelationId::new(), exec_request);
@@ -521,6 +592,13 @@ where
         }) = result
         {
             self.post_state_hash = Some(post_state_hash);
+
+            if let Ok(StoredValue::CLValue(cl_registry)) =
+                self.query(self.post_state_hash, Key::SystemContractRegistry, &[])
+            {
+                let registry = CLValue::into_t::<SystemContractRegistry>(cl_registry).unwrap();
+                self.system_contract_registry = Some(registry);
+            }
         }
 
         self.upgrade_results.push(result);
@@ -643,26 +721,36 @@ where
 
     /// Returns the [`ContractHash`] of the mint, panics if it can't be found.
     pub fn get_mint_contract_hash(&self) -> ContractHash {
-        self.mint_contract_hash
+        self.get_system_contract_hash(MINT)
+            .cloned()
             .expect("Unable to obtain mint contract. Please run genesis first.")
     }
 
     /// Returns the [`ContractHash`] of the "handle payment" contract, panics if it can't be found.
     pub fn get_handle_payment_contract_hash(&self) -> ContractHash {
-        self.handle_payment_contract_hash
+        self.get_system_contract_hash(HANDLE_PAYMENT)
+            .cloned()
             .expect("Unable to obtain handle payment contract. Please run genesis first.")
     }
 
     /// Returns the [`ContractHash`] of the "standard payment" contract, panics if it can't be
     /// found.
     pub fn get_standard_payment_contract_hash(&self) -> ContractHash {
-        self.standard_payment_hash
+        self.get_system_contract_hash(STANDARD_PAYMENT)
+            .cloned()
             .expect("Unable to obtain standard payment contract. Please run genesis first.")
     }
 
+    fn get_system_contract_hash(&self, contract_name: &str) -> Option<&ContractHash> {
+        self.system_contract_registry
+            .as_ref()
+            .and_then(|registry| registry.get(contract_name))
+    }
+
     /// Returns the [`ContractHash`] of the "auction" contract, panics if it can't be found.
     pub fn get_auction_contract_hash(&self) -> ContractHash {
-        self.auction_contract_hash
+        self.get_system_contract_hash(AUCTION)
+            .cloned()
             .expect("Unable to obtain auction contract. Please run genesis first.")
     }
 
@@ -729,7 +817,8 @@ where
     /// Returns the "handle payment" contract, panics if it can't be found.
     pub fn get_handle_payment_contract(&self) -> Contract {
         let handle_payment_contract: Key = self
-            .handle_payment_contract_hash
+            .get_system_contract_hash(HANDLE_PAYMENT)
+            .cloned()
             .expect("should have handle payment contract uref")
             .into();
         self.query(None, handle_payment_contract, &[])
@@ -864,7 +953,7 @@ where
         utils::get_exec_costs(exec_results)
     }
 
-    /// Returns the `Gas` const of the last exec.
+    /// Returns the `Gas` cost of the last exec.
     pub fn last_exec_gas_cost(&self) -> Gas {
         let exec_results = self
             .exec_results
@@ -874,6 +963,26 @@ where
         exec_result.cost()
     }
 
+    /// Returns the result of the last exec.
+    pub fn last_exec_result(&self) -> &ExecutionResult {
+        let exec_results = self
+            .exec_results
+            .last()
+            .expect("Expected to be called after run()");
+        exec_results.get(0).expect("should have result").as_ref()
+    }
+
+    /// Assert that last error is the expected one.
+    ///
+    /// NOTE: we're using string-based representation for checking equality
+    /// as the `Error` type does not implement `Eq` (many of its subvariants don't).
+    pub fn assert_error(&self, expected_error: Error) {
+        match self.get_error() {
+            Some(error) => assert_eq!(format!("{:?}", expected_error), format!("{:?}", error)),
+            None => panic!("expected error ({:?}) got success", expected_error),
+        }
+    }
+
     /// Returns the error message of the last exec.
     pub fn exec_error_message(&self, index: usize) -> Option<String> {
         let response = self.get_exec_result(index)?;
@@ -885,8 +994,12 @@ where
         let correlation_id = CorrelationId::new();
         let state_hash = self.get_post_state_hash();
         let request = GetEraValidatorsRequest::new(state_hash, *DEFAULT_PROTOCOL_VERSION);
+        let system_contract_registry = self
+            .system_contract_registry
+            .clone()
+            .expect("System contract registry not found. Please run genesis first.");
         self.engine_state
-            .get_era_validators(correlation_id, request)
+            .get_era_validators(correlation_id, Some(system_contract_registry), request)
             .expect("get era validators should not error")
     }
 
@@ -1037,4 +1150,11 @@ where
         self.transforms = Vec::new();
         self
     }
+
+    /// Returns a trie by hash.
+    pub fn get_trie(&mut self, state_hash: Digest) -> Option<Trie<Key, StoredValue>> {
+        self.engine_state
+            .get_trie(CorrelationId::default(), state_hash)
+            .unwrap()
+    }
 }
